//  written by Zheng Gengbin
//  support task parallelism
//  and other features defined in HPF2.0

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#if STDC_HEADERS
#include <stdlib.h>
#include <string.h>
#else
#include <strings.h>
#endif

#ifdef HAVE_MALLOC_H
#include <malloc.h>
#endif

#include "compilerApi.h"
#include "anls_att.h"

extern void make_sub_grid(SgStatement *s, SgSymbol *subgrp, Processor_def *pdef, SgArrayRefExp *procs, int way);


/****************************************************************************
                         !HPF$ ON
****************************************************************************/

class on_stack {
SgExpression *exp[10];
int top;
public:
on_stack() { top=0; }

void push(SgExpression *con) {
	exp[top] = con;
	top++;
     }

SgExpression * pop() {
	top --;
	return exp[top]; 
     }

void cleanup() { top = 0; }

int is_empty() {
	return top==0?1:0;
     }

int number() { return top; }

SgExpression * look(int i) { 
	ASSERT(i<top);
  	return exp[i]; 
     }
};

class on_stack onstack_ptr;
class on_stack taskstack_ptr;

void analyse_resident_clause(SgStatement *s, SgExpression *op)
{
   ASSERT(op->variant() == RESIDENT_OP);
   for (SgExpression *exprlist = op->lhs(); exprlist; exprlist=exprlist->rhs())
   {
	SgExpression *expr = exprlist->lhs();
	SgSymbol *sym = expr->symbol();
	sym->addAttribute(Resident_Def, 0, 0);
        s->addAttribute(Resident_Def, sym, sizeof(sym));
   }
}

void analyse_on_stmt(SgStatement *s)
{
     // ON clause
   SgExpression *expr = s->expr(0);
   switch (expr->variant()) {
   case HOME_PROCS: {
     }
     break;
   case HOME_OP: {
     }
     break;
   }

     // RESIDENT clause
   SgExpression *res_op = s->expr(1);
   if (res_op) analyse_resident_clause(s, res_op);
}


void trans_on_stmt(SgStatement *s)
{
   SgExpression *op = s->expr(0);
   SgExpression *expr = op->lhs();
   SgExpression *condition = NULL;
   switch (op->variant()) {
   case HOME_PROCS: {
	SgSymbol *procname = expr->symbol();
	if (procname->haveAttribute(PROCESSORS_BIT)) {
	    Processor_def *pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);
	    ASSERT(pdef);
	    SgArrayRefExp *procs = isSgArrayRefExp(expr);
	    if (procs) {
	       SgSymbol *subgrp = s->getSymbolInCurrentScope(mixstring, "subgrp", procname->identifier(), expr->id());
	       make_sub_grid(s, subgrp, pdef, procs, BEFORE);
	       condition = new SgVarRefExp(subgrp);
	    }
	    else {
	       SgSymbol *grp = s->getSymbolInCurrentScope(mixstring, "grp", procname->identifier());
	       condition = new SgVarRefExp(grp);
	    }
	}
	else {	 	// must be Template
	    printf("Sorry, cannot support Template right now.\n");
	    ASSERT(0);
	}
     }
     break;
   case HOME_OP: {
	SgSymbol *name = expr->symbol();
	SgArrayRefExp *array = isSgArrayRefExp(expr);
	if (array) {
	    SgSymbol *sdd = s->getSymbolInCurrentScope(mixstring, "sdd", name->identifier(),expr->id());
	    SgSymbol *dad = s->getSymbolInCurrentScope(mixstring, "dad", name->identifier());
	    int dim = array->nonScalarDim();
	    SgStatement *sect = make_new_array_section(s, sdd, dim , dad);
	    s->insertStmtBefore(*sect, *s->controlParent());
	    define_array_from_arrayref(s, array, sdd);
	    condition = new SgVarRefExp(sdd);
	}
	else {
	    SgSymbol *dad = s->getSymbolInCurrentScope(mixstring, "dad", name->identifier());
	    condition = new SgVarRefExp(dad);
	}
     }
     break;
   }

   if (taskstack_ptr.is_empty())
   {
      if (condition)
	onstack_ptr.push(condition);
   }
   else
     // change to if pcrc_on()
   if (condition) {
	SgStatement *dir_end = s->lastNodeOfStmt();
	SgExpression *con_exp = make_on(s, condition);
	s->thebif->variant = LOGIF_NODE;
	s->setExpression(0, con_exp);
	s->setExpression(1, NULL);
	s->setExpression(2, NULL);
	dir_end->thebif->variant = CONTROL_END;
   }
}

int need_wrap_on_clause()
{
   return onstack_ptr.is_empty()?0:1;
}

SgStatement * wrap_by_on_clause(SgStatement *s, SgStatement *inner)
{
   SgStatement *wrap = inner;
   int number = onstack_ptr.number();
   for (int i = 0 ; i< number ; i++) {
	SgExpression *on_exp = make_on(s, onstack_ptr.look(i));
        wrap  = new SgIfStmt(*on_exp, *wrap);
   }
  return wrap;
}

/****************************************************************************
                         !HPF$ RESIDENT
****************************************************************************/


void analyse_resident_stmt(SgStatement *s)
{
   SgExpression *resident_op = s->expr(0);
   ASSERT(resident_op);
   analyse_resident_clause(s, resident_op);
}



void trans_resident_stmt(SgStatement *s)
{
   SgExpression *resident_op = s->expr(0);
   ASSERT(resident_op);
   analyse_resident_clause(s, resident_op);
}


/****************************************************************************
			!HPF$ TASK_REGION
****************************************************************************/

void analyse_task_stmt(SgStatement *s)
{
	taskstack_ptr.push(NULL);
}


/****************************************************************************
                         !HPF$ END
****************************************************************************/

void analyse_dirend_stmt(SgStatement *s)
{
   SgStatement *cp = s->controlParent();
     // remove all Resident_Def attributes
   int i;
   for (i=0; i<cp->numberOfAttributes(Resident_Def); i++) 
   {
	SgSymbol *sym = (SgSymbol *)cp->attributeValue(i, Resident_Def);
	sym->deleteAllAttribute(Resident_Def);
   }
}

void trans_dirend_stmt(SgStatement *s)
{
   SgStatement *cp = s->controlParent();
     // pop ON stack
   if (cp->variant() == BLOCK_ON_STMT)
   {
	onstack_ptr.pop();
   }
   if (cp->variant() == TASK_STMT)
   {
	taskstack_ptr.pop();
   }
     // remove all Resident_Def attributes
   int i;
   for (i=0; i<cp->numberOfAttributes(Resident_Def); i++) 
   {
	SgSymbol *sym = (SgSymbol *)cp->attributeValue(i, Resident_Def);
	sym->deleteAllAttribute(Resident_Def);
   }
}

/****************************************************************************
			!HPF$ IO_PROCESSOR
****************************************************************************/

void trans_io_processor_stmt(SgStatement *s)
{
   SgExpression *exprlist = s->expr(0);
   SgArrayRefExp *proc = isSgArrayRefExp(exprlist->lhs());
   ASSERT(proc);
   SgSymbol *procname = proc->symbol();
   Processor_def *pdef = (Processor_def *)procname->attributeValue(0, Proc_Def);
   ASSERT(pdef);
   SgSymbol *subgrp = s->getSymbolInCurrentScope(mixstring, "subgrp", procname->identifier(), proc->id());
   make_sub_grid(s, subgrp, pdef, proc, BEFORE);
   SgStatement *io_proc = make_set_io_process(s, new SgVarRefExp(subgrp));
   s->insertStmtBefore(*io_proc, *s->controlParent());
   s->deleteStmt();
}

